/*
Beregost Crash Fixer
Copyright (C) 2008-2010  Raphael Wong

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package asc64.beregost;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.io.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.filechooser.FileFilter;
import javax.swing.text.*;
import java.text.MessageFormat;
import java.util.*;
import java.util.zip.*;

public class BeregostGUI implements ActionListener {

  //declare and initialize constants
  public static final String VERSION = "1.9";
  static final String newline = "\n";
  private static final long ACTOR_OFFSET_POS = 0x54;
  private static final long ACTOR_NUM_POS = 0x58;
  private static final long ACTOR_BLOCK_SIZE = 0x110;

  JDialog aboutDialog;
  JFrame frame;
  JTextArea log;
  JScrollPane scrollPane;
  JButton openFile, processFile, aboutOK;
  JFileChooser fileChooser;
  JMenu fileMenu, langMenu, helpMenu;
  JMenuItem exit, instructions, about, langEN, langES, langDE, langFR, langPL, langRU, langZHS;
  JProgressBar progressBar;
  JTextField openFileText;
  JLabel openLabel;
  JLabel[] aboutLabel;
  RandomAccessFile file;
  ProcessSAV process;
  ResourceBundle text;
  String inst;
  MessageFormat mf = new MessageFormat("");

  //constructor
  BeregostGUI() {

    //create the main window
    frame = new JFrame();
    frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    frame.setResizable(false);

    //create the menu bar
    JMenuBar menuBar = new JMenuBar();

    fileMenu = new JMenu();
    fileMenu.setMnemonic(KeyEvent.VK_F);
    exit = new JMenuItem();
    exit.addActionListener(this);
    fileMenu.add(exit);
    menuBar.add(fileMenu);

    langMenu = new JMenu();
    langMenu.setMnemonic(KeyEvent.VK_L);
    langEN = new JMenuItem("English");
    langEN.addActionListener(this);
    langMenu.add(langEN);
    langES = new JMenuItem("Castiliano");
    langES.addActionListener(this);
    langMenu.add(langES);
    langDE = new JMenuItem("Deutsch");
    langDE.addActionListener(this);
    langMenu.add(langDE);
    langFR = new JMenuItem("Fran\u00E7ais");
    langFR.addActionListener(this);
    langMenu.add(langFR);
    langPL = new JMenuItem("Polski");
    langPL.addActionListener(this);
    langMenu.add(langPL);
    langRU = new JMenuItem("\u0420\u0443\u0441\u0441\u043A\u0438\u0439");
    langRU.addActionListener(this);
    langMenu.add(langRU);
    langZHS = new JMenuItem("\u7B80\u4F53\u4E2D\u6587");
    langZHS.addActionListener(this);
    langMenu.add(langZHS);
    menuBar.add(langMenu);

    helpMenu = new JMenu();
    helpMenu.setMnemonic(KeyEvent.VK_H);
    instructions = new JMenuItem();
    instructions.addActionListener(this);
    instructions.setAccelerator(KeyStroke.getKeyStroke("F1"));
    helpMenu.add(instructions);
    about = new JMenuItem();
    about.addActionListener(this);
    helpMenu.add(about);
    menuBar.add(helpMenu);

    frame.setJMenuBar(menuBar);


    //create the about dialog
    aboutDialog = new JDialog(frame, true);

    JPanel aboutPane = new JPanel();
    aboutPane.setLayout(new BoxLayout(aboutPane, BoxLayout.PAGE_AXIS));

    aboutLabel = new JLabel[4];

    for (int i = 0; i < aboutLabel.length; i++) {
      if (i == 2) {
        aboutPane.add(Box.createRigidArea(new Dimension(0, 5)));
        aboutPane.add(new JSeparator(JSeparator.HORIZONTAL));
        aboutPane.add(Box.createRigidArea(new Dimension(0, 5)));
      }
      aboutLabel[i] = new JLabel();
      aboutLabel[i].setAlignmentX(Component.CENTER_ALIGNMENT);
      aboutPane.add(aboutLabel[i]);
    }

    aboutPane.add(Box.createRigidArea(new Dimension(0, 5)));
    aboutPane.add(new JSeparator(JSeparator.HORIZONTAL));
    aboutPane.add(Box.createRigidArea(new Dimension(0, 5)));

    aboutOK = new JButton();
    aboutOK.setAlignmentX(Component.CENTER_ALIGNMENT);
    aboutOK.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        aboutDialog.setVisible(false);
        aboutDialog.setLocationRelativeTo(frame);
      }
    });
    aboutPane.add(aboutOK);
    aboutPane.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));

    aboutDialog.setContentPane(aboutPane);
    aboutDialog.setResizable(false);
    aboutDialog.getRootPane().setDefaultButton(aboutOK);

    //create content pane for main window
    JPanel contentPane = new JPanel();
    contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.PAGE_AXIS));
    contentPane.setOpaque(true);
    frame.setContentPane(contentPane);

    //create the panel that selects the file
    JPanel openPanel = new JPanel(new FlowLayout());
    openPanel.setBorder(BorderFactory.createEmptyBorder(0, 0, 10, 0));

    openLabel = new JLabel();
    openPanel.add(openLabel);

    openFileText = new JTextField(40);
    openFileText.setEditable(false);
    openPanel.add(openFileText);

    openFile = new JButton("...");
    openFile.setPreferredSize(new Dimension(20, 20));
    openFile.addActionListener(new BeregostGUI.OpenFileActionListener());
    openPanel.add(openFile);

    contentPane.add(openPanel);

    //create the progress bar
    progressBar = new JProgressBar();
    progressBar.setStringPainted(true);
    progressBar.setString("");

    contentPane.add(progressBar);

    //create a log with scroll pane
    log = new JTextArea(8, 50);
    log.setEditable(false);
    log.setLineWrap(true);
    log.setWrapStyleWord(true);
    scrollPane = new JScrollPane(log);
    scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);

    contentPane.add(scrollPane);

    //create the panel with the process button
    JPanel processPanel = new JPanel();

    processFile = new JButton();
    processFile.setMnemonic(KeyEvent.VK_P);
    processFile.addActionListener(new BeregostGUI.ProcessFileActionListener());
    processFile.setEnabled(false);
    processPanel.add(processFile);

    contentPane.add(processPanel);

    //initialise text
    initText(Locale.ENGLISH);

    //create the file chooser
    fileChooser = new JFileChooser();
    fileChooser.setAcceptAllFileFilterUsed(false);
    fileChooser.addChoosableFileFilter(new SAVFilter());

    //finalize windows and show main frame
    frame.setLocationRelativeTo(null);
    frame.setVisible(true);
  }
  
  //fetch all the text into target HashMap
  private void initText(Locale l) {
    text = ResourceBundle.getBundle("asc64/beregost/Text", l);

    frame.setTitle(text.getString("frame"));
    fileMenu.setText(text.getString("fileMenu"));
    exit.setText(text.getString("exit"));
    langMenu.setText(text.getString("langMenu"));
    helpMenu.setText(text.getString("helpMenu"));
    instructions.setText(text.getString("instructions"));
    about.setText(text.getString("about"));
    aboutLabel[0].setText(text.getString("about0"));
    aboutLabel[1].setText(mf.format(text.getString("about1"), VERSION));
    aboutLabel[2].setText(text.getString("about2"));
    aboutLabel[3].setText(text.getString("about3"));
    aboutOK.setText(text.getString("aboutOK"));
    openLabel.setText(text.getString("openLabel"));
    processFile.setText(text.getString("processFile"));
    
    String instPattern = "{0}\n{1}\n{2}\n{3}\n\n{4}";
    String[] instFormat = {
      text.getString("inst0"),
      text.getString("inst1"),
      text.getString("inst2"),
      text.getString("inst3"),
      text.getString("inst4"),
    };
    mf.applyPattern(instPattern);
    inst = mf.format(instFormat);

    aboutDialog.pack();
    frame.pack();

    return;
  }

  //instance method to verify that the loaded file is valid
  private String verifyFile(File f) throws IOException {
    try  {
      file = new RandomAccessFile(f, "rw");
    } catch (FileNotFoundException fnf) {
      return text.getString("verifyUnable");
    }

    if (BeregostUtils.getExtension(f).equals(BeregostUtils.SAV_FILE)) {
      if (!BeregostUtils.readAscii(file, 0).matches("SAV V1.0")) {
        return text.getString("verifyInvalidHeader");
      } else {
        return null;
      }
    }

    return text.getString("verifyIncorrect");

  }

  //reset all the components that are displaying a status
  private void resetGUI() {
    try {
      Document document = log.getDocument();
      document.remove(0, document.getLength());
    } catch (BadLocationException bl) {
    }
    progressBar.setValue(0);
    progressBar.setString("");
    openFileText.setText(fileChooser.getSelectedFile().getAbsolutePath());
  }

  //action listener for the menu items
  public void actionPerformed(ActionEvent e) {
    JMenuItem sourceMenu = (JMenuItem)e.getSource();
    if (sourceMenu == exit) {
      System.exit(0);
    } else if (sourceMenu == instructions) {
      try {
        Document document = log.getDocument();
        document.remove(0, document.getLength());
      } catch (BadLocationException bl) {
      }
      log.append(inst);
    } else if (sourceMenu == about) {
      aboutDialog.setLocationRelativeTo(frame);
      aboutDialog.setVisible(true);
    } else if (sourceMenu == langEN) {
      initText(Locale.ENGLISH);
    } else if (sourceMenu == langES) {
      initText(new Locale("es"));
    } else if (sourceMenu == langDE) {
      initText(Locale.GERMAN);
    } else if (sourceMenu == langFR) {
      initText(Locale.FRENCH);
    } else if (sourceMenu == langPL) {
      initText(new Locale("pl"));
    } else if (sourceMenu == langRU) {
      initText(new Locale("ru"));
    } else if (sourceMenu == langZHS) {
      initText(new Locale("zhs")); //Chinese (Simplified)
    }
  }

  //inner class action listener for the open file button
  private class OpenFileActionListener implements ActionListener {
    public void actionPerformed(ActionEvent e) {
      int intValue = fileChooser.showOpenDialog(null);
      if (intValue == JFileChooser.APPROVE_OPTION) {

        resetGUI();
        //check to see if the selected file is a valid file
        try {
          String verify = verifyFile(fileChooser.getSelectedFile());
          if (verify == null) {
            log.append(mf.format(text.getString("verifyLoaded"), fileChooser.getSelectedFile().getAbsolutePath()) +newline);
            instructions.setEnabled(false);
            processFile.setEnabled(true);
          } else {
            log.append(mf.format(text.getString("verifyInvalidFile"), fileChooser.getSelectedFile().getAbsolutePath()) + " (" +verify+ ")" +newline);
            file.close();
          }
        } catch (IOException io) {
        }
      }
   }
  }
////inner class action listener for the process file button////////////////////////////////////
  private class ProcessFileActionListener implements ActionListener {
    public void actionPerformed(ActionEvent e) {
      if (BeregostUtils.getExtension(fileChooser.getSelectedFile()).equals(BeregostUtils.SAV_FILE)) {
        process = new ProcessSAV();
        process.execute();
        processFile.setEnabled(false);
        openFile.setEnabled(false);
      }
    }
  }

////inner class area file filter///////////////////////////////////////////////////////////////
  private class SAVFilter extends FileFilter {

    public boolean accept(File f) {
      if (f.isDirectory()) {
        return true;
      }

      String ext = BeregostUtils.getExtension(f);
      if (ext.equals(BeregostUtils.SAV_FILE)) {
        return true;
      } else {
        return false;
      }
    }

    public String getDescription() {
      return text.getString("desc");
    }
  }

////inner class SwingWorker object for processing ARE files/////////////////////////////////////////
  private class ProcessARE{
    boolean errorInARE = false;

    public void run() throws IOException, DataFormatException {
      SwingUtilities.invokeLater(new Runnable() {
        public void run() {
          log.append(mf.format(text.getString("processing"), process.filename));
        }
      });

      int actorOffset = BeregostUtils.readInt(file, ACTOR_OFFSET_POS);
      short actorNumber = BeregostUtils.readShort(file, ACTOR_NUM_POS);

      for (int i = 0; i < actorNumber; i++) {
        final String actorName = BeregostUtils.readAscii(file, actorOffset + i * ACTOR_BLOCK_SIZE + 0x80);
        int actorCreOffset = BeregostUtils.readInt(file, actorOffset + i * ACTOR_BLOCK_SIZE + 0x88);
        if (actorCreOffset == 0) {
          throw new DataFormatException();
        }
        int actorCreSpellMemOff = BeregostUtils.readInt(file, actorCreOffset + 0x2a8);
        int actorCreSpellMemNum = BeregostUtils.readInt(file, actorCreOffset + 0x2ac);

        for (int j = 0; j < actorCreSpellMemNum; j++) {
          final short lvl = BeregostUtils.readShort(file, actorCreOffset + actorCreSpellMemOff + j * 0x10);
          final short num1 = BeregostUtils.readShort(file, actorCreOffset + actorCreSpellMemOff + j * 0x10 + 0x2);
          final short num2 = BeregostUtils.readShort(file, actorCreOffset + actorCreSpellMemOff + j * 0x10 + 0x4);
          final short type = BeregostUtils.readShort(file, actorCreOffset + actorCreSpellMemOff + j * 0x10 + 0x6);
          final int a = i + 1;
          final int b = j;
          if ((lvl < 0 || lvl > 8) || (num1 < 0 || num1 > 9) || (num2 < 0 || num2 > 9) || (type < 0 || type > 2)) {
            errorInARE = true;
            SwingUtilities.invokeLater(new Runnable() {
              public void run() {
                log.append(newline 
                         + text.getString("error0") + ": " +a+ " [" +actorName.trim()+ "], " + text.getString("error1") + ": " +b+ "." +newline
                         //+ "Corrupted hex values (normal range)" + ": " +Integer.toHexString((int)lvl).toUpperCase()+ " (0-8) " +Integer.toHexString((int)num1).toUpperCase()+ " (0-9) "
                         //+ Integer.toHexString((int)num2).toUpperCase()+ " (0-9) " +Integer.toHexString((int)type).toUpperCase()+ " (0-2)" +newline
                         + text.getString("error2") +newline);
              }
            });
            BeregostUtils.writeLong(file, actorCreOffset + actorCreSpellMemOff + j * 0x10, 0);
          }
        }
      }

      if (!errorInARE) {
        SwingUtilities.invokeLater(new Runnable() {
          public void run() {
            log.append(text.getString("noerror") +newline);
          }
        });
      }

    }

  }

////SwingWorker subclass for processing SAV files///////////////////////////////////////////////////
  private class ProcessSAV extends SwingWorker<Void, String> {
    String filename;
    
    //write an entry into the output SAV file
    void writeSAVEntry(RandomAccessFile sav, File f) throws IOException {
      int dataLengthUncomp;
      int dataLengthComp;
      byte[] dataComp = new byte[(int)f.length() * 2];

      if (f.length() == 0) {
        byte[] dataCompEmpty = new byte[]{0x78, 0x01, 0x03, 0x00, 0x00, 0x00, 0x00, 0x01};
        dataLengthUncomp = 0;
        dataLengthComp = 8;
        dataComp = Arrays.copyOf(dataCompEmpty, dataLengthComp);
      } else {
        BufferedInputStream bis = new BufferedInputStream(new FileInputStream(f));
        byte[] dataUncomp = new byte[(int)f.length()];
        dataLengthUncomp = bis.read(dataUncomp);
        bis.close();

        Deflater deflater = new Deflater(Deflater.BEST_SPEED);
        deflater.setInput(dataUncomp);
        deflater.finish();
        dataLengthComp = deflater.deflate(dataComp);
        dataComp = Arrays.copyOf(dataComp, dataLengthComp);
        deflater.end();
      }

      //write the file entry into the output SAV file
      long offset = sav.getFilePointer();
      BeregostUtils.writeInt(sav, offset, (int)(f.getName().length() + 1));
      sav.writeBytes(f.getName());
      sav.writeByte(0);
      BeregostUtils.writeInt(sav, sav.getFilePointer(), dataLengthUncomp);
      BeregostUtils.writeInt(sav, sav.getFilePointer(), dataLengthComp);
      sav.write(dataComp, 0, dataLengthComp);
      
      f.delete();
    }

    protected Void doInBackground() throws IOException {
      ProcessARE processArea = null;

      //make a new reference to the sav file so we can make the current file 'file', saving duplication of code
      RandomAccessFile SAVFile = file;
      File outputDirectory = new File(fileChooser.getCurrentDirectory(), "_" + fileChooser.getCurrentDirectory().getName());
      File outputFile =  new File(fileChooser.getCurrentDirectory(), "$BALDUR");
      RandomAccessFile outputSAV = new RandomAccessFile(outputFile, "rw");

      try {
        outputDirectory.mkdir();
        long pos = 8;
        while (true) {
          //read all information for one file entry
          int filenameLength = BeregostUtils.readInt(SAVFile, pos);
          filename = BeregostUtils.readAscii(SAVFile, pos+4, filenameLength - 1);
          int dataLengthUncomp = BeregostUtils.readInt(SAVFile, pos+4+filenameLength);
          int dataLengthComp = BeregostUtils.readInt(SAVFile, pos+4+filenameLength+4);
          byte[] dataComp = BeregostUtils.readByte(SAVFile, pos+4+filenameLength+4+4, dataLengthComp);

          //create an output file for the file entry
          setProgress((int)((pos * 100 / SAVFile.length())));
          publish(filename);
          File outputFileEntry = new File(outputDirectory, filename);
          BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(outputFileEntry));

          //create the ZLIB inflater, then decompress
          Inflater inflater = new Inflater();
          inflater.setInput(dataComp);
          byte[] dataUncomp = new byte[dataLengthUncomp];
          inflater.inflate(dataUncomp);
          inflater.end();

          //convert byte[] to char[] array, then write to file
          bos.write(dataUncomp, 0, dataUncomp.length);
          
          //close the output stream so that we can open a RandomAccessFile stream on the file entry
          bos.close();

          file = new RandomAccessFile(outputFileEntry, "rw");
          //process only if filename is AR6700.ARE, FW3300.ARE, or AR3300.ARE
          if ((filename.compareToIgnoreCase("ar6700.are") == 0) ||
              (filename.compareToIgnoreCase("fw3300.are") == 0) ||
              (filename.compareToIgnoreCase("ar3300.are") == 0)) {
            processArea = new ProcessARE();
            processArea.run();
          }
          file.close();

          //increment the position to be at the start of the next file entry
          pos += 4+filenameLength+4+4+dataLengthComp;
          if (pos >= SAVFile.length()) {
            setProgress(100);
            break;
          }
        }

        //reconstruct the SAV file
        log.append(text.getString("reconstructing") +newline);
        outputSAV.writeBytes("SAV V1.0");

        File files[] = outputDirectory.listFiles();
        Arrays.sort(files);
        int index = 0;
        for (File currentFile : files) {
          writeSAVEntry(outputSAV, currentFile);
        }

      } catch (IOException io) {
          SwingUtilities.invokeLater(new Runnable() {
            public void run() {
              log.append(text.getString("errorProcessing") + ": " +fileChooser.getSelectedFile().getAbsoluteFile()+ ". (" + text.getString("ioError") + ")" +newline);
            }
          });
      } catch (DataFormatException df) {
          //DFE is thrown in this program if null values are present in the input, indicating an end of file problem
          //ProcessARE also throws this when the CRE offset header is 0
          SwingUtilities.invokeLater(new Runnable() {
            public void run() {
              log.append(text.getString("errorProcessing") + ": " +fileChooser.getSelectedFile().getAbsoluteFile()+ ". (" + text.getString("errorNonBeregost") + ")" +newline);
            }
          });
      } finally {
        file.close();
        SAVFile.close();
        outputSAV.close();
        
        outputDirectory.delete();

        File backup = new File(fileChooser.getCurrentDirectory()+ "/BALDUR.BAK");
        if (backup.exists()) {
          String options[] = {text.getString("overwrite"), text.getString("cancel")};
          if (JOptionPane.showOptionDialog(frame, mf.format(text.getString("backup"), backup.getName()), "", JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE, null, options, options[0]) == 0) {
            backup.delete();
            fileChooser.getSelectedFile().getAbsoluteFile().renameTo(backup);
            fileChooser.getSelectedFile().getAbsoluteFile().delete();
            outputFile.renameTo(fileChooser.getSelectedFile().getAbsoluteFile());
            log.append(mf.format(text.getString("overwriting"), backup.getAbsoluteFile()) + newline);
          } else {
            outputFile.delete();
            log.append(text.getString("cancelProcess") +newline);
          }
        } else {
          fileChooser.getSelectedFile().getAbsoluteFile().renameTo(backup);
          fileChooser.getSelectedFile().getAbsoluteFile().delete();
          outputFile.renameTo(fileChooser.getSelectedFile().getAbsoluteFile());
          outputFile.delete();
        }
      }

    return null;
    }

    protected void process(java.util.List<String> chunks) {
      if (!this.isDone()) {
        progressBar.setValue(process.getProgress());
        progressBar.setString(mf.format(text.getString("scanning"), chunks.get(chunks.size() - 1)));
      }
    }

    protected void done() {
      progressBar.setValue(100);
      progressBar.setString(text.getString("done"));
      openFile.setEnabled(true);
      instructions.setEnabled(true);
      openFileText.setText(null);
      log.append(text.getString("complete"));
    }
  }

////Property change listener for ProcessSAV///////////////////////////////////////////////////////////////
//unused now due to use of publish() and process()////////////////////////////////////////////////////////
//  private class ProcessSAVListener implements PropertyChangeListener {
//    public void propertyChange(PropertyChangeEvent evt) {
//      String property = evt.getPropertyName();
//      if ("progress" == property && progressBar.getValue() != 100) {
//        progressBar.setString(null);
//        progressBar.setValue(process.getProgress());
//      }
//    }
//  }
  
}